(* DelphiXE2-Unit (Unicode-Version)
   verschiedene Unterroutinen
   ==========================
   - String-Konvertierungen ANSI <-> OEM
   - div. String-Routinen (z.B. Leerezeichen anfgen oder lschen)
   - Befehlszeilen auswerten
   - Umwandlung Grad:Min:Sek -> String und zurck
   - Konvertierung von Hex, Oktal und Binr <-> string
   - Substrings und Zahlen bis Begrenzer extrahieren
   - Gltigkeitsprfung fr Zahlen
   - Umwandlung Integer und Double -> String mit fhrenden Nullen und Tausendermark.
   - Dateigre als String
   - Zusammensetzen von Word und Cardinal aus Bytes
   - Typdeklarationen fr das Zerlegen von Word, LongWort und Int64 in Bytes und Words
   - schnelles Sortieren einer Liste (Quicksort)
   - Standarddialoge: Info, Fehler und Besttigung
   - Absolute Position ermitteln
   - Verwaltung von History-Listen
   - Text-Filter

    Dr. J. Rathlev, D-24222 Schwentinental (info(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Neue Zusammenstellung - Mai. 2007
   Sprachabhngige Teile in UnitStrings - Dez. 2011
   letzte nderung - Apr. 2013
   *)

unit JrUtils;

interface

uses
  Winapi.Windows, System.SysUtils, System.Classes, Vcl.Graphics,
  Vcl.Forms, Vcl.Controls, Vcl.ExtCtrls, Vcl.Printers, Vcl.ComCtrls,
  Vcl.Dialogs, Messages, IniFiles;

type
  DateStr = string[10];
  TCvTable = array [0..127] of AnsiChar;

  TSetArray = array [1..32] of byte;
  PSetArray = ^TSetArray;

  // Compare function for Quicksort
  TCompFunction = function (const arg1,arg2) : boolean of object;

const
  ChrID = '#';             // Kennzeichner fr num. Zeichenwerte
  Minus = '-';
  Komma = ',';
  Punkt = '.';
  DPunkt = ':';
  Semikolon = ';';
  DecPt = '.';
  VertBar = '|';
  Quote = '"';
  Space = ' ';
  Tab = #9;
  CrLf = #13#10;
  Cr = #13;
  Lf = #10;

  FloatChars : set of char = ['0'..'9',DecPt,'-','+','E','e'];

  MaxInt64 = High(int64);

  // Bildschirm-Auflsung bei der Programmentwicklung
  PixelsPerInchOnDesign = 96;
  { "Scaled = true" passt die Formulare automatisch an andere Textgren an
    Fr die Berechnung von Spaltenbreiten, o.. muss dann zustzlich folgende
    Umrechnung verwendet werden:
    n:=MulDiv(n(96),Screen.PixelsPerInch,PixelsPerInchOnDesign)
  }

  xDef = 1.0;              // Standardwert ReadNxtDbl
  nDef = 1;                // Standardwert ReadNxtInt
  cDef = #0;               // Standardwert ReadNxtChr

  TabToANSI : TCvTable =(
    #$C7,#$FC,#$E9,#$E2,#$E4,#$E0,#$E5,#$E7,#$EA,#$EB,#$E9,#$E8,#$EF,#$EE,#$C4,#$C5,
    #$C9,#$E6,#$C6,#$F4,#$F6,#$F2,#$FB,#$F9,#$FF,#$D6,#$DC,#$A2,#$A3,#$A5,#$20,#$20,
    #$E1,#$ED,#$F3,#$FA,#$F1,#$D1,#$61,#$6F,#$BF,#$20,#$AC,#$BD,#$BC,#$A1,#$AB,#$BB,
    #$20,#$20,#$20,#$A6,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$AD,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$DF,#$B6,#$20,#$20,#$20,#$B5,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$B1,#$20,#$20,#$20,#$20,#$F7,#$20,#$B0,#$20,#$B7,#$20,#$B3,#$B2,#$20,#$20);
  TabToOEM : TCvTable =(
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$AD,#$9B,#$9C,#$20,#$9D,#$B3,#$15,#$20,#$20,#$20,#$AE,#$AA,#$C4,#$20,#$20,
    #$F8,#$F1,#$FD,#$FC,#$20,#$E6,#$E3,#$20,#$20,#$20,#$20,#$AF,#$AC,#$AB,#$20,#$A8,
    #$41,#$41,#$41,#$41,#$8E,#$8F,#$92,#$80,#$45,#$90,#$45,#$45,#$49,#$49,#$49,#$49,
    #$44,#$A5,#$4F,#$4F,#$4F,#$4F,#$99,#$20,#$20,#$55,#$55,#$55,#$9A,#$59,#$20,#$E1,
    #$85,#$61,#$83,#$61,#$84,#$86,#$91,#$87,#$8A,#$82,#$88,#$89,#$8D,#$A1,#$8C,#$8B,
    #$20,#$A4,#$95,#$A2,#$93,#$6F,#$94,#$F6,#$20,#$97,#$A3,#$96,#$81,#$79,#$20,#$98);

  UpperLetters : TSysCharSet = ['A'..'Z'];
  LowerLetters : TSysCharSet = ['a'..'z'];
  Digits : TSysCharSet = ['0'..'9'];
  HexDigits : TSysCharSet = ['0'..'9','A'..'F','a'..'f'];
  AlphaNum : TSysCharSet = ['0'..'9','A'..'Z','a'..'z'];

  CenterPos : TPoint = (X : -1; Y : -1);
  MaxHist : integer = 50;

type
  TBoolFunction = function : boolean of object;
  TIntegerFunction = function : integer;

  TWinkelFormat = (fmDezGrad,fmDezMin,fmSekunden);
  TTextChange = (tcNone,tcLower,tcUpper);

  TWord = record
    case integer of
    0 : (Both : word);
    1 : (Lo,Hi : byte);
    end;

  TLongWord = record
    case integer of
    0 : (LongWord : cardinal);
    1 : (Lo,Hi : word);
    2 : (LoL,LoH,HiL,HiH : byte);
    3 : (Bytes : array [0..3] of Byte);
    end;

  TInt64 = record
    case integer of
    0: (AsInt64 : int64);
    1: (Lo, Hi  : Cardinal);
    2: (Cardinals: array [0..1] of Cardinal);
    3: (Words: array [0..3] of Word);
    4: (Bytes: array [0..7] of Byte);
    end;

  TFontStyleToByte = record
    case integer of
    1 : (Style : TFontStyles);
    2 : (Value : byte);
    end;

  TArea = record
    case integer of
    0 : (Left,Top,Width,Height: integer);
    1 : (TopLeft,WidthHeight: TPoint);
    end;

  TFPoint = record
    X,Y : double;
    end;

  TFRect = record
    case integer of
    0 : (Left,Top,Right,Bottom : double);
    1 : (TopLeft,BottomRight : TFPoint);
    end;

{ ---------------------------------------------------------------- }
(* Verzgerungszeit unter Verwendung der Windows Waitable Timer Objects
   Wait: warten bis Zeit in ms abgelaufen ist (min. Zeit 100 ns) *)

  TWaitTimer = class (TObject)
    FTimer   : THandle;
    constructor Create;
    destructor Destroy;  override;
    procedure Wait (DelayTime : double);
    end;

{ ---------------------------------------------------------------- }
// Anzeige eines Hinweisfenster (THintWindow), das nach einstellbarer Zeit (Delay)
// automatisch verschwindet
  TTimerHint = class (THintWindow)
  private
    FTimer : TTimer;
    FOnTerminate : TNotifyEvent;
    procedure Terminate (Sender : TObject);
  protected
    procedure CreateParams(var Params: TCreateParams); override;
  public
    constructor Create (AOwner: TComponent; Delay : integer);  // Delay in ms
    destructor Destroy; override;
    procedure ShowHint (r : TRect; AHint : string);
    procedure HideHint;
    property OnTerminate : TNotifyEvent read FOnTerminate write FOnTerminate;
    end;

{ ---------------------------------------------------------------- }
// Umwandeln eines DOS-Zeichens in eines Grobuchstaben (auch Umlaute)
function OEMUpCase (c : AnsiChar) : AnsiChar;

// Umwandeln eines ANSI-Zeichens in eines Grobuchstaben (auch Umlaute)
function ANSIUpCase (c : AnsiChar) : AnsiChar;

// Umwandeln eines DOS-Strings in Grobuchstaben (auch Umlaute)
function OEMUpString (const S : AnsiString) : AnsiString;

// Umwandeln eines ANSI-Strings in Grobuchstaben (auch Umlaute)
function ANSIUpString (const S : AnsiString) : AnsiString;

// Umwandeln von ISO-8859-Zeichen fr eine Sortierung
function ANSISortConvert (const S : string) : string;
function ANSISortUpper (const S : string) : string;

// Umwandeln eines Strings in Grobuchstaben (Umlaute in AE, OE, UE) zu Sortierzwecken
function OEMUpSort (const S : AnsiString) : AnsiString;

// Umwandeln eines OEM-Strings nach ANSI
function StrToAnsi (const s : AnsiString) : AnsiString;

// Umwandeln eines ANSI-Strings nach OEM
function StrToOEM (const s : AnsiString) : AnsiString;

// Umwandeln eines Zeichens in Kleinbuchstaben  (fehlt in Unit System)
function LowCase(Ch: WideChar): WideChar;

{ ---------------------------------------------------------------- }
// Substring-Pos. ohne Unterschreidung gro/klein
function TextPos (const Substr, S: string): Integer;
function TextPosEx (const Substr, S: string; Offset : cardinal): Integer;

// Substring-Pos. von rechts ohne Unterscheidung gro/klein
function RightTextPos(const Substr,s : string) : integer;
function RightTextPosEx (const Substr, S: string; Offset : cardinal): Integer;

// prfen, ob Text bestimmte Zeichen enthlt
function ContainsCharacters (const AText : string; AChars : array of Char) : boolean;

// Zeichen aus Text entfernen
function RemoveCharacters (const AText : AnsiString; AChars : array of AnsiChar) : AnsiString; overload;
function RemoveCharacters (const AText : string; AChars : array of Char) : string; overload;

// Text auf Gro- oder Kleinschreibung ndern
function TextChangeCase(const s: string; CaseMode : TTextChange) : string;

// Info-String aus zwei Zeilen erzeugen
function MakeInfoStr (const s1,s2 : string) : string;

// erster Buchstabe gro
function SetInitial(const s : string) : string;

// Formatierung ohne Exception bei Fehler
function TryFormat(const AFormat: string; const Args: array of const): string;

{ ---------------------------------------------------------------- }
// String mit Leerstellen erzeugen
function FillSpace (len : integer) : string;

// fhrende Leerstellen ergnzen
function AddSp (const S : string;
                len     : integer) : string;

// Leerstellen am Ende ergnzen
function ExtSp (const S : string;
                len   : integer) : string;

// fhrende Leerstellen lschen
function RemSp (const S : string) : string;

// Leerstellen am Ende lschen
function CutSp (const S : string) : string;

// alle Leerstellen im String lschen
function DelSp (const S : string) : string;

// mehrfache Leerstellen im String lschen
function DelMulSp (const S : string) : string;

// Steuerzeichen im String lschen bzw. ersetzen
function DelCtrlChars(const S : string) : string;
function CrLfToChar (s : string; c : char) : string;
function CharToCr (s : string; c : char) : string;

// erstes nichtleeres Zeichen suchen
function SeekNoSp (const s : string) : integer;

// nchstes Leerzeichen oder Tab suchen
function SeekSpace (const s : string) : integer;

// String mit Zeichen erzeugen
function FillStr (c   : char;
                  len : integer) : string;

// fhrende Zeichen ergnzen
function AddChar (const S : string;
                  c       : char;
                  len     : integer) : string;

// Zeichen am Ende ergnzen
function ExtChar (const S : string;
                  c       : char;
                  len     : integer) : string;

// ein Zeichen durch ein anderes ersetzen
function ReplChars (const s         : AnsiString;
                    OldChar,NewChar : AnsiChar) : AnsiString; overload;
function ReplChars (const s         : string;
                    OldChar,NewChar : char) : string; overload;

// fhrende Zeichen lschen
function RemChar (const S : string; c : char) : string;

// Zeichen am Ende lschen
function CutChar (const S : string; c : char) : string;

// Zeichen im String lschen
function DelChar (const S : AnsiString; c : AnsiChar) : AnsiString; overload;
function DelChar (const S : string; c : char) : string; overload;

// Leerzeichen gegen _ und / gegen % austauschen oder zurck
function ReplOptionChar (s : string; Back : boolean) : string;

// Zeichen prfen
function IsLetter (c : char) : boolean;
function IsDigit (c : char) : boolean;
function IsHexDigit (c : char) : boolean;

// spez. Zeichen in Strings durch Hexzahl (Unicode) ersetzen und umgekehrt
function SpecCharToValue(const ss,sr : string) : string;
function ValueToSpecChar(const FName : string) : string;

// String auf feste Lnge bringen (krzen oder mit Leerzeichen auffllen)
function FixLength (const S : string;
                    len     : integer) : string;

// Teilstring lschen als Funktion
function DeleteSubString(const s : string; Index,Count : integer)  : string;

// String am Anfang lschen bis ausschlielich "c"
function DeleteUntilChar(const s : string; c : char) : string;

// Anzahl von SubStr in S zhlen
function CountSubStr (const ASubStr,AString : string) : integer;

// Replace characters in "AText" from "Index1" to "Index2" with string "AToText"
function ReplaceChars (const AText,AToText : string; Index1,Index2 : cardinal) : string;

{ --------------------------------------------------------------- }
// Check if AText matches filters: AFilter="<filter1><sep><filter2><sep>.."
function MatchesFilter(const AText,AFilter : string; Sep : char) : boolean;

{ ---------------------------------------------------------------- }
// Dateifilter-Index ermitteln (siehe TOpenDialog)
function GetFilterIndex(AFilter,AExtension : string) : integer;

{ ---------------------------------------------------------------- }
// Routinen zur Auswertung einer Befehlszeile
// prfe, ob die ersten Zeichen einer Option mit dem Parameter bereinstimmen
function CompareOption (const Param,Option : string) : boolean;

// Option vom Typ /option:value einlesen
function ReadOptionValue (var Param : string; const Option : string) : boolean;

// Steuerdatei mit Startoptionen einlesen
function ReadOptionFile (FName : string) : string;

{ ---------------------------------------------------------------- }
(* Integer, Double oder String aus einem String s bis zum nchsten
   Begrenzer (Del) lesen
   s wird um den verarbeiteten Teil gekrzt *)
function ReadNxtInt (var s   : String;
                     Del     : char;
                     Default : int64;
                     var err : boolean) : int64; overload;

function ReadNxtInt (var s   : String;
                     Del     : char;
                     Default : int64) : int64; overload;

function ReadNxtDbl (var s   : String;
                     Del     : char;
                     Default : double;
                     var err : boolean) : double; overload;

function ReadNxtDbl (var s   : String;
                     Del     : char;
                     Default : double) : double; overload;

function ReadNxtStr (var s   : AnsiString;
                     Del     : AnsiChar) : AnsiString; overload;
function ReadNxtStr (var s   : AnsiString;
                     Del     : AnsiChar;
                     Default : AnsiString) : AnsiString; overload;
function ReadNxtStr (var s   : String;
                     Del     : char) : string; overload;
function ReadNxtStr (var s   : String;
                     Del     : char;
                     Default : string) : string; overload;

function TryReadNxtInt (s : string; Del : char; var Value : integer) : boolean;

function GetNxtStr (const s   : String; Del : char) : string;
function GetTrimStr (const S : string; Del : char) : string;
function GetNxtInt (const S : string; Del : char; Default : int64) : int64;

function ReadNxtWord (var s : string;
                      Dels  : TSysCharSet) : string;

function ReadNxtChr (var s   : String;
                     Del     : char;
                     Default : char;
                     var err : boolean) : char; overload;

function ReadNxtChr (var s   : String;
                     Del     : char;
                     Default : char) : char; overload;

function ReadNxtQuotedStr (var s : string;
                           ADelim,AQuote : char) : string;

function MakeQuotedStr (const s : string; CheckChars : array of char) : string;

function CommaTextToLines (Commatext : string) : string;

function MaxTextWidth(const Text : string; Canvas : TCanvas) : integer;

function CountLines (const Text : string) : integer;

// duplicate all quote characters (e..g &, ")
function DuplicateQuotes(const S: string; Quote : Char): string;

// Position of char in string starting at "Offset", ignore parts between quotes
function PosChar (const s: string; AChar,AQuote : char; Offset: Cardinal): Integer;

function GetPluralString (const sNo,sOne,sMany : string; n : integer) : string; overload;
function GetPluralString (const sNo,sOne,sMany : string; n : integer; const s : string) : string; overload;

{ ---------------------------------------------------------------- }
(* Testen, ob eine gltige Integer oder Double Zahl in einem String s
   bis zum nchsten Begrenzer (Del) steht *)
function TestNxtNumber (s       : String;
                        Del     : char ) : boolean;

// Wert auf Gltigkeit prfen - Result=true : x <> NoVal 
function Valid (x,NoVal : double) : boolean;

{ ---------------------------------------------------------------- }
// Dezimaltrenner setzen und lesen, Standard = Punkt 
procedure SetDecPtChar (Zeichen : char);
function GetDecPtChar : char;

// Tausendermarkierung, Standard = Leerzeichen 
procedure ThMarkChar (Zeichen : char);

// Integer-Zahl in String umsetzen 
function StrInt (x : int64;
                 n : integer) : string;

// Integer-Zahl in String mit fhrenden Nullen umsetzen 
function ZStrInt (x : int64;
                  n : integer) : string;

// Stellenanzahl einer Integerzahl
function NumberOfDigits (Value : int64) : integer;

// Integer-Zahl als Hex-Zahl, Digits = Mindestanzahl der Stellen,
//      Group = Gruppierung von Stellen
function IntToHexStr (Value,Digits,Group : cardinal) : string;

// Integer-Zahl als Dezimal-Zahl, Digits = Mindestanzahl der Stellen,
//      Group = Gruppierung von Stellen
function IntToDecimal(Value : int64; Digits,Group : cardinal) : string; overload;

// Integer-Zahl als Oktal-Zahl, Digits = Mindestanzahl der Stellen,
//      Group = Gruppierung von Stellen
function IntToOctal (Value,Digits,Group : cardinal) : string;

// Integer-Zahl als Binr-Zahl, Digits = Mindestanzahl der Stellen
//      Group = Gruppierung von Stellen
function IntToBin (Value,Digits,Group : cardinal) : string;

// Konvertiere String als Hexwert (ohne $)
function TryHexStrToInt64(S: string; var val : int64) : boolean;
function TryHexStrToInt(const S : string; var val : integer) : boolean;
function HexStrToInt (const S : string) : int64;

// Konvertiere String als Oktalwert
function TryOctalStrToInt64(S: string; var val : int64) : boolean;
function TryOctalStrToInt(const S : string; var val : integer) : boolean;
function OctalStrToInt (const S : string) : int64;

// Konvertiere String als Binrwert
function TryBinStrToInt64(S: string; var val : int64) : boolean;
function TryBinStrToInt(const S : string; var val : integer) : boolean;
function BinStrToInt (const S : string): int64;

// Double-Zahl in String umsetzen
function StrDbl (x   : double;
                 n,d : integer) : string; overload;
function StrDbl (x   : double;
                 n,d : integer; DecPt : char) : string; overload;

// Double-Zahl in String mit fhrenden Nullen umsetzen
function ZStrDbl (x   : double;
                  n,d : integer) : string; overload;
function ZStrDbl (x   : double;
                  n,d : integer; DecPt : char) : string; overload;

// Double-Zahl in String in Dreiergruppen umsetzen
function NStrDbl (x : double;
                  n,d : integer) : string; overload;
function NStrDbl (x   : double;
                  n,d : integer; DecPt : char) : string; overload;

// Konvertierung String <-> Fliekomma-Zahl mit Punkt als Dezimaltrenner
function StringToFloat (const s : string; Default : extended) : extended;
function FloatToString (Value : extended; Format: TFloatFormat; Precision, Digits: Integer) : string;

// Bytes, je nach Gre in String umsetzen
function SizeToStr (Bytes : int64) : string; overload;
function SizeToStr (Bytes : int64; DecPt : char) : string; overload;

// String in Dateigre wandeln
function StrToSize (s : string; var size : int64) : boolean;

// Umwandeln einer Menge in einen Hex-String und zurck
function SetToHex (ASet : PSetArray; Low,High : integer) : string;
procedure HexToSet (const S : string; ASet: PSetArray);

{ ---------------------------------------------------------------- }
// Dezimalkomma ersetzen
function ReplDecPt (s : string) : string;

// Dezimalpunkt einsetzen 
function SetDecPt (s : string) : string;

(* alle aktuellen Dezimaltrenner durch Punkt ersetzen *)
function ReplaceDecSeps (const s : string) : string;

(* alle Punkte durch aktuellen Dezimlatrenner ersetzen *)
function SetDecSeps (const s : string) : string;
{ ---------------------------------------------------------------- }
// Cardinal in Hi- oder Lo-Wort zerlegen
function HiWord (n : cardinal) : word;

function LoWord (n : cardinal) : word;

// Word aus Hi- und Lo-Byte zusammensetzen
function BytesToWord (HiByte,LoByte: byte) : word;

// Cardinal aus Hi- und Lo-Wort zusammensetzen
function WordsToCardinal (HiWord,LoWord : word) : cardinal;

// Cardinal aus 4 Bytes zusammensetzen
function BytesToCardinal (HHByte,HLByte,LHByte,LLByte : word) : cardinal;

// Int64 aus Hi- und Lo-Cardinal zusammensetzen
function CardinalToInt64 (HiCard,LoCard : cardinal) : int64;

// Convert 16-bit oder 8-Bit number to hex string
function HexByte (b : byte) : AnsiString;

function HexWord (w : word) : AnsiString;

{ ---------------------------------------------------------------- }
// schnelles Sortieren einer Liste
procedure QuickSort (var SList;
                     Count,RSize : Integer;
                     Compare     : TCompFunction);

{ ---------------------------------------------------------------- }
// erweiterte Drucker-Angaben (alle Angaben in mm)
function GetPaperWidth (APrinter : TPrinter) : integer;
function GetPaperHeight (APrinter : TPrinter) : integer;
function GetLeftOffset (APrinter : TPrinter) : integer;
function GetTopOffset (APrinter : TPrinter) : integer;
function GetMaxWidth (APrinter : TPrinter) : integer;
function GetMaxHeight (APrinter : TPrinter) : integer;

// Duplex-Druck
function SupportsDuplex (APrinter : TPrinter) : Boolean;
function UsesDuplex (APrinter : TPrinter) : Boolean;
procedure SetToDuplex (APrinter : TPrinter);

{ ---------------------------------------------------------------- }
// Prfen, ob ein Fenster auf den Bildschirm passt
procedure CheckScreenBounds (AScreen         : TScreen;
                             var ALeft,ATop : integer;
                             AWidth,AHeight : integer);
procedure FitToScreen (AScreen : TScreen; Control : TControl);

// Position einer Form an den Bildschirm anpassen
procedure AdjustFormPosition (AScreen : TScreen; AForm : TForm;
          APos : TPoint; AtBottom : boolean = false);

{ ---------------------------------------------------------------- }
// MessageDlg in Bildschirmmitte (X<0) oder an Position X,Y
function MessageDialog(const Title,Msg: string; DlgType: TMsgDlgType;
                       Buttons: TMsgDlgButtons; DefaultButton : TMsgDlgBtn;
                       Pos : TPoint; Delay : integer;
                       ADefaultMonitor : TDefaultMonitor = dmActiveForm) : integer; overload;
function MessageDialog(const Title,Msg: string; DlgType: TMsgDlgType;
                       Buttons: TMsgDlgButtons;
                       Pos : TPoint; Delay : integer;
                       ADefaultMonitor : TDefaultMonitor = dmActiveForm) : integer; overload;
function MessageDialog(const Title,Msg: string; DlgType: TMsgDlgType;
                       Buttons: TMsgDlgButtons) : integer; overload;
function MessageDialog(const Msg: string; DlgType: TMsgDlgType;
                       Buttons: TMsgDlgButtons) : integer; overload;
function MessageDialog(Pos : TPoint; const Msg: string; DlgType: TMsgDlgType;
                       Buttons: TMsgDlgButtons) : integer;  overload;

function ConfirmDialog (const Title,Msg : string; Pos : TPoint) : boolean; overload;
function ConfirmDialog (const Title,Msg : string) : boolean; overload;
function ConfirmDialog (const Msg : string; DefaultButton : TMsgDlgBtn = mbYes) : boolean; overload;
function ConfirmDialog (Pos : TPoint; const Msg : string; DefaultButton : TMsgDlgBtn = mbYes;
                        ADefaultMonitor : TDefaultMonitor = dmActiveForm) : boolean; overload;

procedure InfoDialog (const Title,Msg : string; Pos : TPoint); overload;
procedure InfoDialog (const Title,Msg : string; Delay : integer); overload;
procedure InfoDialog (const Title,Msg : string); overload;
procedure InfoDialog (const Msg : string); overload;
procedure InfoDialog (Pos : TPoint; const Msg : string); overload;

procedure ErrorDialog (const Title,Msg : string; Pos : TPoint); overload;
procedure ErrorDialog (const Title,Msg : string; x,y : integer); overload;
procedure ErrorDialog (const Title,Msg : string; Delay : integer); overload;
procedure ErrorDialog (const Title,Msg : string); overload;
procedure ErrorDialog (const Msg : string); overload;
procedure ErrorDialog (Pos : TPoint; const Msg : string); overload;

{ ---------------------------------------------------------------- }
// get current cursor position
function CursorPos : TPoint; overload;
function CursorPos (Offset : TPoint): TPoint; overload;
function AddOffsetPos(Pos1,Pos2 : TPoint) : TPoint;

// position of component
function TopLeftPos (AControl : TControl) : TPoint; overload;
function TopLeftPos (AControl : TControl; Offset : TPoint) : TPoint; overload;
function BottomLeftPos (AControl : TControl) : TPoint; overload;
function BottomLeftPos (AControl : TControl; Offset : TPoint) : TPoint; overload;
function TopRightPos (AControl : TControl) : TPoint; overload;
function TopRightPos (AControl : TControl; Offset : TPoint) : TPoint; overload;
function BottomRightPos (AControl : TControl) : TPoint; overload;
function BottomRightPos (AControl : TControl; Offset : TPoint) : TPoint; overload;

// area of component
function GetRect (AControl : TControl) : TRect;

{ ---------------------------------------------------------------- }
// History list management
procedure LoadHistory (IniFile : TIniFile; Section,Ident : string;
                       History : TStrings; MaxCount : integer); overload;
procedure LoadHistory (IniFile : TIniFile; Section,Ident : string;
                       History : TStrings); overload;
procedure LoadHistory (IniName : string; Section,Ident : string;
                       History : TStrings; MaxCount : integer); overload;
procedure LoadHistory (IniName : string; Section,Ident : string;
                       History : TStrings); overload;

procedure SaveHistory (IniFile : TIniFile; Section,Ident : string;
                       Erase : boolean; History : TStrings; MaxCount : integer); overload;
procedure SaveHistory (IniFile : TIniFile; Section,Ident : string;
                       Erase : boolean; History : TStrings); overload;
procedure SaveHistory (IniName: string; Section,Ident : string;
                       Erase : boolean; History : TStrings; MaxCount : integer); overload;
procedure SaveHistory (IniName: string; Section,Ident : string;
                       Erase : boolean; History : TStrings); overload;

procedure AddToHistory (History : TStrings; hs : string; MaxCount : integer); overload;
procedure AddToHistory (History : TStrings; hs : string); overload;
procedure RemoveFromHistory (History : TStrings; hs : string);

{ ---------------------------------------------------------------- }
// Entferne alle Objekte einer String-Liste oder einer ListView-Liste aus dem Speicher
procedure FreeListObjects(Liste : TStrings);
procedure FreeListViewData (Liste : TListItems);

{ ---------------------------------------------------------------- }
// Listview-Index aus Caption ermitteln (wie IndexOf bei TListBox)
function GetListViewIndex (lv : TListView; ACaption : string): integer;

// Subitem-Index aus der Mausposition ermitteln (nur vsReport)
function GetColumnIndexAt (ListView : TListView; Pos : integer) : integer;

// TopItem auf Index setzen (nur vsReport)
procedure SetListViewTopItem (lv : TListView; AIndex : integer; Select : boolean);

{ ---------------------------------------------------------------- }
// Tastaturpuffer lschen
function ClearKeyboardBuffer : Integer;

{ ---------------------------------------------------------------- }
var
  DezTr  : char;   // Dezimalzeichen
  ThMark : char;   // Zeichen fr Tausendergruppen
  TmMark : char;   // Zeichen fr Uhrzeitgruppen hh:mm:ss

{ ---------------------------------------------------------------- }
implementation

uses StrUtils, Masks, DateUtils, WinApiUtils, WinShell, ShlObj, UnitStrings,
  Math, WinSpool;

{ ---------------------------------------------------------------- }
(* Umwandeln eines Zeichens in eines Grobuchstaben (auch Umlaute)
   OEM-Zeichen *)
function OEMUpCase (c : AnsiChar) : AnsiChar;
begin
  if c<#127 then Result:=upcase(c)
  else begin
    case c of
    '' : Result:='';
    '' : Result:='';
    '' : Result:='';
    '' : Result:='S';
      else Result:=c;
      end;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines Zeichens in eines Grobuchstaben (auch Umlaute)
   ANSI-Zeichen *)
function ANSIUpCase (c : AnsiChar) : AnsiChar;
begin
  if c<#224 then Result:=upcase(c)
  else Result:=AnsiChar(ord(c)-32);
{  begin
    case c of
    '' : Result:='';
    '' : Result:='';
    '' : Result:='';
    '' : Result:='S';
      else Result:=c;
      end;
    end;          }
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines Strings in Grobuchstaben (auch Umlaute) *)
function OEMUpString (const S : AnsiString) : AnsiString;
var
  i  : integer;
  ns : AnsiString;
  c  : AnsiChar;
begin
  ns:='';
  for i:=1 to length(s) do begin
    c:=s[i];
    if c<#127 then ns:=ns+upcase(c)
    else begin
      case c of
      '' : ns:=ns+'';
      '' : ns:=ns+'';
      '' : ns:=ns+'';
      '' : ns:=ns+'SS';
        else ns:=ns+c;
        end;
      end;
    end;
  OEMUpString:=ns;
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines ANSI-Strings in Grobuchstaben (auch Umlaute) *)
function ANSIUpString (const S : AnsiString) : AnsiString;
var
  i  : integer;
  ns : AnsiString;
  c  : AnsiChar;
begin
  ns:='';
  for i:=1 to length(s) do begin
    c:=s[i];
    if c<#127 then ns:=ns+upcase(c)
    else begin
      case c of
      '' : ns:=ns+'';
      '' : ns:=ns+'';
      '' : ns:=ns+'';
      '' : ns:=ns+'SS';
        else ns:=ns+c;
        end;
      end;
    end;
  ANSIUpString:=ns;
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln von ISO-8859-Zeichen fr eine Sortierung *)
function ANSISortConvert (const S : string) : string;
var
  i  : integer;
  ns : string;
  c  : char;
begin
  Result:='';
  for i:=1 to length(s) do begin
    c:=s[i];
    if c<#127 then Result:=Result+c
    else begin
      case c of
      '' : Result:=Result+'Ae';
      '' : Result:=Result+'Oe';
      '' : Result:=Result+'Ue';
      '' : Result:=Result+'ae';
      '' : Result:=Result+'oe';
      '' : Result:=Result+'ue';
      '' : Result:=Result+'ss';
      #$C6 : Result:=Result+#$5B; // nach 'Z'  - dnisch
      #$D8 : Result:=Result+#$5C; // nach 'Z'
      #$C5 : Result:=Result+#$5D; // nach 'Z'
      #$E6 : Result:=Result+#$7B; // nach 'z'
      #$F8 : Result:=Result+#$7C; // nach 'z'
      #$E5 : Result:=Result+#$7D; // nach 'z'
      #$C0..#$C3 : Result:=Result+'A';
      #$C7 : Result:=Result+'C';
      #$C8..#$CB : Result:=Result+'E';
      #$CC..#$CF : Result:=Result+'I';
      #$D0 : Result:=Result+'D';
      #$D1 : Result:=Result+'N';
      #$D2..#$D5 : Result:=Result+'O';
      #$D9..#$DB :  Result:=Result+'U';
      #$DD : Result:=Result+'Y';
      #$E0..#$E3 : Result:=Result+'a';
      #$E7 : Result:=Result+'c';
      #$E8..#$EB : Result:=Result+'e';
      #$EC..#$EF : Result:=Result+'i';
      #$F0 : Result:=Result+'d';
      #$F1 : Result:=Result+'n';
      #$F2..#$F5 : Result:=Result+'o';
      #$F9..#$FB :  Result:=Result+'u';
      #$FD,#$FF : Result:=Result+'y';
      else Result:=Result+c;
        end;
      end;
    end;
  end;

(* Umwandeln von ISO-8859-Zeichen fr eine Sortierung mit Grobuchastaben*)
function ANSISortUpper (const S : string) : string;
begin
  Result:=ANSISortConvert(AnsiUpperCase(s));
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines Strings in Grobuchstaben (Umlaute in AE, OE, UE)
   zu Sortierzwecken *)
function OEMUpSort (const S : AnsiString) : AnsiString;
var
  i  : integer;
  ns : AnsiString;
  c  : AnsiChar;
begin
  ns:='';
  for i:=1 to length(s) do begin
    c:=s[i];
    if c<#127 then ns:=ns+upcase(c)
    else begin
      case c of
      '','' : ns:=ns+'AE';
      '','' : ns:=ns+'OE';
      '','' : ns:=ns+'UE';
      ''     : ns:=ns+'SS';
        else ns:=ns+c;
        end;
      end;
    end;
  OEMUpSort:=ns;
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines OEM-Strings nach ANSI *)
function StrToAnsi (const s : AnsiString) : AnsiString;
var
  i  : integer;
begin
  Result:=s;
  for i:=1 to length(Result) do begin
    if Result[i]>=#128 then Result[i]:=TabToANSI[ord(Result[i])-128]
    else if Result[i]=#$15 then Result[i]:=#$A7;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines ANSI-Strings nach OEM *)
function StrToOEM (const s : AnsiString) : AnsiString;
var
  i  : integer;
begin
  Result:=s;
  for i:=1 to length(Result) do begin
    if Result[i]>=#128 then Result[i]:=TabToOEM[ord(Result[i])-128];
    end;
  end;

{ ---------------------------------------------------------------- }
// Umwandeln eines Zeichens in Kleinbuchstaben  (fehlt in Unit System)
function LowCase(Ch: WideChar): WideChar;
begin
  Result:=Ch;
  if (Ch>='A') and (Ch<='Z') then inc(Result,Ord('a') - Ord('A'));
  end;

{------------------------------------------------------------------}
(* Substring-Pos. ohne Unterschreidung gro/klein *)
function TextPos (const Substr, S: string): Integer;
begin
  Result:=Pos (AnsiLowerCase(Substr),AnsiLowerCase(s));
  end;

function TextPosEx (const Substr, S: string; Offset : cardinal): Integer;
begin
  Result:=PosEx (AnsiLowerCase(Substr),AnsiLowerCase(s),Offset);
  end;

(* Substring-Pos. von rechts ohne Unterscheidung gro/klein *)
function RightTextPos(const Substr,s : string) : integer;
begin
  Result:=RightTextPosEx(Substr,s,length(s));
  end;

function RightTextPosEx (const Substr, S: string; Offset : cardinal): Integer;
begin
  Result:=0;
  if length(s)>0 then begin
    Result:=Offset;
    while not AnsiSameText(Substr,copy(s,Result,length(Substr))) and (Result>0) do dec(Result)
    end;
  end;

function ContainsCharacters (const AText : string; AChars : array of Char) : boolean;
var
  i : integer;
begin
  Result:=true;
  for i:=Low(AChars) to High(AChars) do if TextPos(AChars[i],AText)>0 then Exit;
  Result:=false;
  end;

// Zeichen aus Text entfernen
function RemoveCharacters (const AText : AnsiString; AChars : array of AnsiChar) : AnsiString;
var
  i : integer;
begin
  Result:=AText;
  for i:=Low(AChars) to High(AChars) do Result:=AnsiReplaceStr(Result,AChars[i],'');
  end;

function RemoveCharacters (const AText : string; AChars : array of Char) : string;
var
  i : integer;
begin
  Result:=AText;
  for i:=Low(AChars) to High(AChars) do Result:=AnsiReplaceStr(Result,AChars[i],'');
  end;

function SetInitial(const s : string) : string;
begin
  Result:=s;
  Result[1]:=UpCase(Result[1]);
  end;

// Formatierung ohne Exception bei Fehler
function TryFormat(const AFormat: string; const Args: array of const): string;
begin
  try
    Result:=Format(AFormat,Args);
  except
    Result:=AFormat;
    end;
  end;

{ --------------------------------------------------------------- }
// Text auf Gro- oder Kleinschreibung ndern
function TextChangeCase(const s: string; CaseMode : TTextChange) : string;
begin
  if CaseMode=tcLower then Result:=AnsiLowercase(s)
  else if CaseMode=tcUpper then Result:=AnsiUppercase(s)
  else Result:=s;
  end;

{ --------------------------------------------------------------- }
// Info-String aus zwei Zeilen erzeugen
function MakeInfoStr (const s1,s2 : string) : string;
begin
  Result:=StringReplace(s1,'&','&&',[rfReplaceAll])+sLineBreak+
          StringReplace(s2,'&','&&',[rfReplaceAll]);
  end;

{------------------------------------------------------------------}
(* String mit Leerstellen erzeugen *)
function FillSpace (len : integer) : string;
var
  i : byte;
  s : string;
begin
  s:='';
  for i:=1 to len do s:=s+' ';
  Result:=s;
  end;

{------------------------------------------------------------------}
(* fhrende Leerstellen ergnzen *)
function AddSp (const S : string;
                len     : integer) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=succ(length(Result)) to len do Result:=' '+Result;
  end;

{------------------------------------------------------------------}
(* Leerstellen am Ende ergnzen *)
function ExtSp (const S : string;
                len     : integer) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=succ(length(Result)) to len do Result:=Result+' ';
  end;

{------------------------------------------------------------------}
(* fhrende Leerstellen und Tabs lschen *)
function RemSp (const S : string) : string;
var
  i : integer;
begin
  Result:=s; i:=1;
  while (i<=length(Result)) and ((Result[i]=Space) or (Result[i]=Tab)) do inc(i);
  delete(Result,1,pred(i));
  end;

{ --------------------------------------------------------------- }
(* Leerstellen am Ende lschen *)
function CutSp (const S : string) : string;
var
  i : integer;
begin
  Result:=s; i:=length(Result);
  while (i>0) and (Result[i]=Space) do dec(i);
  delete(Result,succ(i),length(Result));
  end;

{ --------------------------------------------------------------- }
(* alle Leerstellen im String lschen *)
function DelSp (const S : string) : string;
var
  i  : integer;
begin
  Result:=s; i:=1;
  while (i<=length(Result)) do begin
    if Result[i]=Space then delete(Result,i,1) else inc(i);
    end;
  end;

{ --------------------------------------------------------------- }
(* mehrfache Leerstellen im String lschen *)
function DelMulSp (const S : string) : string;
var
  i  : integer;
begin
  Result:=s;
  if length(Result)>0 then begin
    i:=1;
    while (i<length(Result)) do begin
      if (Result[i]=Space) and (Result[i+1]=Space) then delete(Result,i,1) else inc(i);
      end;
    if (Result[i]=Space) then delete(Result,i,1);
    end;
  end;

{ --------------------------------------------------------------- }
(* Steuerzeichen im String lschen *)
function DelCtrlChars(const S : string) : string;
var
  i  : integer;
begin
  Result:=s;
  if length(Result)>0 then begin
    i:=1;
    while (i<=length(Result)) do begin
      if Result[i]<#32 then delete(Result,i,1) else inc(i);
      end;
    end;
  end;

// Cr-Lf im String durch spez. Zeichen ersetzen
function CrLfToChar (s : string; c : char) : string;
begin
  Result:=DelCtrlChars(AnsiReplaceStr(s,#$0D,c));
  end;

// spez. Zeichen im String durch Cr-Lf ersetzen
function CharToCr (s : string; c : char) : string;
begin
  Result:=AnsiReplaceStr(s,c,#$0D);
  end;

{ --------------------------------------------------------------- }
(* erstes nichtleeres Zeichen suchen *)
function SeekNoSp (const s : string) : integer;
var
  i  : integer;
begin
  i:=1;
  while (i<=length(s)) and ((s[i]=Space) or (s[i]=Tab)) do inc(i);
  if i<=length(s) then SeekNoSp:=i else SeekNoSp:=0;
  end;

{ --------------------------------------------------------------- }
(* Position fr nchstes Leerzeichen oder Tab suchen,
   bei Stringende = Lnge+1 *)
function SeekSpace (const s : string) : integer;
var
  i  : integer;
begin
  if length(s)=0 then Result:=0
  else begin
    i:=1;
    while (i<=length(s)) and (s[i]<>Space) and (s[i]<>Tab) do inc(i);
    Result:=i;
    end;
  end;

{ --------------------------------------------------------------- }
(* String mit Zeichen erzeugen *)
function FillStr (c   : char;
                  len : integer) : string;
var
  i : byte;
  s : string;
begin
  s:='';
  for i:=1 to len do s:=s+c;
  Result:=s;
  end;

{ --------------------------------------------------------------- }
(* fhrende Zeichen ergnzen *)
function AddChar (const S : string;
                  c       : char;
                  len     : integer) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=succ(length(Result)) to len do Result:=c+Result;
  end;

{ --------------------------------------------------------------- }
(* Zeichen am Ende ergnzen *)
function ExtChar (const S : string;
                  c       : char;
                  len     : integer) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=succ(length(Result)) to len do Result:=Result+c;
  end;

{ --------------------------------------------------------------- }
// ein Zeichen durch ein anderes ersetzen
function ReplChars (const s         : AnsiString;
                    OldChar,NewChar : AnsiChar) : AnsiString;
var
  i  : integer;
begin
  Result:=s;
  for i:=1 to length(Result) do if Result[i]=OldChar then Result[i]:=NewChar;
  end;

function ReplChars (const s         : string;
                    OldChar,NewChar : char) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=1 to length(Result) do if Result[i]=OldChar then Result[i]:=NewChar;
  end;

{ --------------------------------------------------------------- }
// fhrende Zeichen lschen
function RemChar (const S : string; c : char) : string;
var
  i : integer;
begin
  Result:=s; i:=1;
  while (i<=length(Result)) and (Result[i]=c) do inc(i);
  delete(Result,1,pred(i));
  end;

{ --------------------------------------------------------------- }
// Zeichen am Ende lschen
function CutChar (const S : string; c : char) : string;
var
  i : integer;
begin
  Result:=s; i:=length(Result);
  while (i>0) and (Result[i]=c) do dec(i);
  delete(Result,succ(i),length(Result));
  end;

{ --------------------------------------------------------------- }
// Zeichen im String lschen
function DelChar (const S : AnsiString; c : AnsiChar) : AnsiString;
var
  i  : integer;
begin
  Result:=s; i:=1;
  while (i<=length(Result)) do begin
    if Result[i]=c then delete(Result,i,1) else inc(i);
    end;
  end;

function DelChar (const S : string; c : char) : string;
var
  i  : integer;
begin
  Result:=s; i:=1;
  while (i<=length(Result)) do begin
    if Result[i]=c then delete(Result,i,1) else inc(i);
    end;
  end;

{ --------------------------------------------------------------- }
(* Leerzeichen gegen _ und / gegen % austauschen oder zurck *)
function ReplOptionChar (s : string; Back : boolean) : string;
var
  j : integer;
begin
  if Back then begin
    for j:=1 to length(s) do begin
      if s[j]='_' then s[j]:=' ';
      if s[j]='%' then s[j]:='/';
      end;
    end
  else begin
    for j:=1 to length(s) do begin
      if s[j]=' ' then s[j]:='_';
      if s[j]='/' then s[j]:='%';
      end;
    end;
  Result:=s;  
  end;

{ --------------------------------------------------------------- }
// Zeichen prfen
function IsLetter (c : char) : boolean;
begin
  Result:=CharInSet(c,UpperLetters) or CharInSet(c,LowerLetters);
  end;

function IsDigit (c : char) : boolean;
begin
  Result:=CharInSet(c,Digits);
  end;

function IsHexDigit (c : char) : boolean;
begin
  Result:=CharInSet(c,HexDigits);
  end;

{ --------------------------------------------------------------- }
// spez. Zeichen in Strings durch Hexzahl (Unicode) ersetzen und umgekehrt
function SpecCharToValue(const ss,sr : string) : string;
var
  i,j  : integer;
  s,sn : string;
begin
  Result:=ss;
  for j:=1 to length(sr) do begin
    s:=''; sn:='%'+IntToHex(word(sr[j]),4);
    for i:=1 to length(Result) do if Result[i]=sr[j] then s:=s+sn else s:=s+Result[i];
    Result:=s;
    end;
  end;

function ValueToSpecChar(const FName : string) : string;
var
  i,j : integer;
begin
  Result:='';
  i:=1;
  while i<=length(FName) do begin
    if FName[i]='%' then begin
      if (i+4<=length(FName)) and TryStrToInt('$'+copy(FName,i+1,4),j) then begin
        Result:=Result+Char(j);
        inc(i,4);
        end
      else Result:=Result+FName[i];
      end
    else Result:=Result+FName[i];
    inc(i);
    end;
  end;

{ --------------------------------------------------------------- }
(* String auf feste Lnge bringen (krzen oder mit Leerzeichen auffllen *)
function FixLength (const S : string;
                    len : integer) : string;
begin
  if length(s)<len then FixLength:=ExtSp (S,len)
  else FixLength:=copy (s,1,len);
  end;

{ --------------------------------------------------------------- }
// Teilstring lschen als Funktion
function DeleteSubString(const s : string; Index,Count : integer)  : string;
begin
  Result:=s;
  Delete(Result,Index,Count);
  end;

{ --------------------------------------------------------------- }
// String am Anfang lschen bis ausschlielich "c", wenn nicht gefunden alles lschen
function DeleteUntilChar(const s : string; c : char) : string;
var
  n : integer;
begin
  Result:=s;
  n:=pos(c,Result);
  if n>0 then Delete(Result,1,n-1) else Result:='';
  end;

{ --------------------------------------------------------------- }
(* Anzahl von SubStr in S zhlen *)
function CountSubStr (const ASubStr,AString : string) : integer;
var
  n : integer;
begin
  Result:=0; n:=1;
  repeat
    n:=PosEx(ASubStr,AString,n+1);
    if n>0 then inc(Result);
    until n=0;
  end;

{ --------------------------------------------------------------- }
// Replace characters in "AText" from "Index1" to "Index2" with string "AToText"
function ReplaceChars (const AText,AToText : string; Index1,Index2 : cardinal) : string;
begin
  Result:=AText;
  Delete (Result,Index1,Index2-Index1+1);
  Insert (AToText,Result,Index1);
  end;

{ --------------------------------------------------------------- }
// Check if AText matches filters: AFilter="<filter1><sep><filter2><sep>.."
function MatchesFilter(const AText,AFilter : string; Sep : char) : boolean;
var
  s : string;
begin
  s:=AFilter;
  repeat
    try
      Result:=MatchesMask(AText,ReadNxtStr(s,Sep));
    except
      Result:=false;
      end;
    until Result or (length(s)=0);
  end;

{ --------------------------------------------------------------- }
// Dateifilter-Index ermitteln (siehe TOpenDialog)
//
function GetFilterIndex(AFilter,AExtension : string) : integer;
var
  n : integer;
begin
  Result:=0; n:=0;
  repeat
    inc(n);
    ReadNxtStr(AFilter,'|');  // Beschreibung berlesen
    if AnsiContainsText(ReadNxtStr(AFilter,'|'),AExtension) then Result:=n;
    until (Result>0) or (length(AFilter)=0);
  if Result=0 then Result:=n;  // letztes Filter (*.*)
  end;

{ ---------------------------------------------------------------- }
// Routinen zur Auswertung einer Befehlszeile
// prfe, ob die ersten Zeichen einer Option mit dem Parameter bereinstimmen
function CompareOption (const Param,Option : string) : boolean;
begin
  Result:=AnsiLowercase(Param)=copy(Option,1,length(Param));
  end;

// Option vom Typ option:value einlesen
function ReadOptionValue (var Param : string; const Option : string) : boolean;
var
  i : integer;
begin
  Result:=false;
  i:=AnsiPos(DPunkt,Param);
  if i=0 then exit;
  if not CompareOption(copy(Param,1,i-1),Option) then exit;
  Delete(Param,1,i); Result:=true;
  end;

// Steuerdatei mit Startoptionen einlesen
function ReadOptionFile (FName : string) : string;
var
  f : TextFile;
  s : string;
begin
  Result:='';
  if FileExists(FName) then begin
    AssignFile(f,FName); Reset(f);
    while not Eof(f) do begin
      readln(f,s);
      s:=Trim(ReadNxtStr(s,';'));  // bis Kommentar ";" lesen
      if length(s)>0 then Result:=Result+'|'+s
      end;
    CloseFile(f);
    if length(Result)>0 then Delete(Result,1,1);
    end
  end;

{------------------------------------------------------------------}
(* Winkel-Wert konvertieren *)
function StrToDeg (s     : string;
                   var v : double) : boolean;
var
 t      : string;
 x,y    : double;
 j,k    : integer;
begin
  j:=0;
  k:=pos(',',s); if k>0 then s[k]:='.';
  if length(s)<>0 then begin
    s:=ReplDecPt(s);
    k:=pos(DPunkt,s);
    if k=0 then k:=pos('/',s);
    if k=0 then k:=pos('"',s);
    if k=0 then val(s,v,j)
    else begin (* Grad:Min *)
      val (copy(s,1,pred(k)),x,j); (* Grad *)
      if j=0 then begin
        t:=copy (s,succ(k),255);
        k:=pos(DPunkt,t);
        if k=0 then k:=pos('/',t);
        if k=0 then k:=pos(#$27,s);
        if k=0 then begin (* dez.Min *)
          if length(t)>0 then begin
            val (t,v,j);
            if (j=0) and (v<60.0) then begin
              if x>=0.0 then v:=x+v/60.0
              else v:=x-v/60.0;
              end
            else j:=1;
            end
          else v:=x;
          end
        else begin (* Sek. *)
          val (copy (t,1,pred(k)),y,j);  (* Min *)
          if (j=0) and (y<60.0) then begin
            t:=copy (t,succ(k),255);
            if length(t)>0 then begin
              val (t,v,j); (* Sek. *)
              if (j=0) and (v<60.0) then begin
                if x>=0.0 then v:=x+(y+v/60.0)/60.0
                else v:=x-(y-v/60.0)/60.0;
                end
              else j:=1;
              end
            else begin
              if x>=0.0 then v:=x+y/60.0
              else v:=x-y/60.0;
              end;
            end
          else j:=1;
          end;
        end;
      end;
    end
  else v:=0.0;
  StrToDeg:=j=0;
  end;

{------------------------------------------------------------------}
(* Winkel in String mit vorgebenem Format umsetzen
   Stellen: w : gesamt
            d : hinter Komma (einschl. Minuten und Sekunden
   Typ : m = fmDezGrad  : Grad,xxx
             fmDezMin   : Grad:Minuten,xx
             fmSekunden : Grad:Minuten:Sekunden,xxx *)
function DegToStr(v      : double;
                  Format : TWinkelFormat;
                  w,d    : integer) : string;
var
  s,t   : string;
  x,y,z : extended;
  i     : integer;
begin
  x:=int(abs(v)); y:=60.0*frac(abs(v));
  case Format of
  fmDezGrad : str(v:w:d,s);
  fmDezMin : begin (* Grad:Min *)
      str (y:(w-6):(d-2),t);
      if copy(t,1,2)='60' then begin
        y:=0.0; x:=x+1.0;
        str (y:(w-6):(d-2),t);
        end;
      if v<0.0 then x:=-x;
      str(x:5:0,s); s:=s+DPunkt+t;
      end;
  fmSekunden : begin (* Grad:Min:Sek *)
      z:=int(y); y:=60.0*frac(y);
      str(y:(w-8):(d-4),t);
      if copy(t,1,2)='60' then begin
        y:=0.0; z:=z+1.0;
        if z>=60.0 then begin
          z:=0.0; x:=x+1.0;
          end;
        str (y:(w-8):(d-4),t);
        end;
      str (z:2:0,s); t:=s+DPunkt+t;
      if v<0.0 then x:=-x;
      str(x:4:0,s); s:=s+':'+t;
      end;
    end;
  i:=pos('.',s); if i>0 then s[i]:=',';
  DegToStr:=s;
  end;

{ ------------------------------------------------------------------- }
(* Integer, Double oder String aus einem String s bis zum nchsten Tenner lesen
   s wird um den verarbeiteten Teil gekrzt *)
function ReadNxtInt (var s   : String;
                     Del     : char;
                     Default : int64;
                     var err : boolean) : int64;
var
  n    : int64;
  i,ic : integer;
begin
  s:=TrimLeft(s); i:=pos(Del,s);
  if i=0 then i:=succ(length(s));
  val(copy(s,1,pred(i)),n,ic);
  if ic=0 then ReadNxtInt:=n
  else begin
    ReadNxtInt:=Default;
    err:=true;
    end;
  delete(s,1,i);
  end;

function ReadNxtInt (var s   : String;
                     Del     : char;
                     Default : int64) : int64;
var
  err : boolean;
begin
  Result:=ReadNxtInt(s,Del,Default,err);
  end;

function TryReadNxtInt (s : string; Del : char; var Value : integer) : boolean;
begin
  Result:=false;
  Value:=ReadNxtInt(s,Del,0,Result);
  Result:=not Result;
  end;

{ ------------------------------------------------------------------- }
function ReadNxtDbl (var s   : String;
                     Del     : char;
                     Default : double;
                     var err : boolean) : double;
var
  i,ic : integer;
  x    : double;
begin
  s:=TrimLeft(s); i:=pos (Del,s);
  if i=0 then i:=succ(length(s));
  val(SetDecPt(copy(s,1,pred(i))),x,ic);
  if ic=0 then ReadNxtDbl:=x
  else begin
    ReadNxtDbl:=Default;
    err:=true;
    end;
  delete(s,1,i);
  end;

function ReadNxtDbl (var s   : String;
                     Del     : char;
                     Default : double) : double;
var
  err : boolean;
begin
  Result:=ReadNxtDbl(s,Del,Default,err);
  end;

{ ------------------------------------------------------------------- }
function ReadNxtStr (var s   : AnsiString;
                     Del     : AnsiChar) : AnsiString;
var
  i : integer;
begin
  if length(s)>0 then begin
    i:=pos (Del,s);
    if i=0 then i:=succ(length(s));
    ReadNxtStr:=copy(s,1,pred(i));
    delete(s,1,i);
    end
  else ReadNxtStr:='';
  end;

function ReadNxtStr (var s   : AnsiString;
                     Del     : AnsiChar;
                     Default : AnsiString) : AnsiString; overload;
var
  t : AnsiString;
begin
  t:=ReadNxtStr(s,Del);
  if length(t)=0 then Result:=Default else Result:=t;
  end;

function ReadNxtStr (var s   : String;
                     Del     : char) : string;
var
  i : integer;
begin
  if length(s)>0 then begin
    i:=pos (Del,s);
    if i=0 then i:=succ(length(s));
    ReadNxtStr:=copy(s,1,pred(i));
    delete(s,1,i);
    end
  else ReadNxtStr:='';
  end;

function ReadNxtStr (var s   : String;
                     Del     : char;
                     Default : string) : string;
var
  t : string;
begin
  t:=ReadNxtStr(s,Del);
  if length(t)=0 then Result:=Default else Result:=t;
  end;

function GetNxtStr (const S   : String; Del : char) : string;
var
  i : integer;
begin
  if length(s)>0 then begin
    i:=pos (Del,s);
    if i=0 then i:=succ(length(s));
    Result:=copy(s,1,pred(i));
    end
  else Result:='';
  end;

function GetTrimStr(const S : string; Del : char) : string;
begin
  Result:=Trim(GetNxtStr(s,Del));
  end;

function GetNxtInt(const S : string; Del : char; Default : int64) : int64;
begin
  if not TryStrToInt64(GetTrimStr(S,Del),Result) then Result:=Default;
  end;

{ ------------------------------------------------------------------- }
(* Wort bis zum nchsten Delimiter lesen *)
function ReadNxtWord (var s : string;
                      Dels  : TSysCharSet) : string;
var
  i,n : integer;
begin
  n:=length(s);
  if n>0 then begin
    i:=1;
    while (i<=n) and not CharInSet(AnsiChar(s[i]),Dels) do inc(i);
    Result:=copy(s,1,pred(i));
    delete(s,1,pred(i));
    end
  else Result:='';
  end;

{ ------------------------------------------------------------------- }
(* ein Zeichen lesen oder num. Wert #nnn in Zeichen umwandeln *)
function ReadNxtChr (var s   : String;
                     Del     : char;
                     Default : char;
                     var err : boolean) : char;
var
  n,i,ic : integer;
begin
  if (length(s)>0) then begin
    i:=pos (Del,s);
    if (s[1]=ChrID) then begin  (* # *)
      if i=0 then i:=succ(length(s));
      val(copy(s,2,pred(i)),n,ic);
      if ic=0 then ReadNxtChr:=chr(n)
      else begin
        ReadNxtChr:=Default; err:=true;
        end;
      end
    else ReadNxtChr:=s[1];
    delete(s,1,i);
    end
  else begin
    ReadNxtChr:=Default; err:=true;
    end;
  end;

function ReadNxtChr (var s   : String;
                     Del     : char;
                     Default : char): char;
var
  err : boolean;
begin
  Result:=ReadNxtChr(s,Del,Default,err);
  end;

function ReadNxtQuotedStr (var s : string; ADelim,AQuote : char) : string;
var
  n : integer;
  qs,fin : boolean;
begin
  s:=TrimLeft(s);
  if length(s)>0 then begin
    if s[1]=AQuote then begin
      n:=2; qs:=true; fin:=false;
      repeat
        if s[n]=AQuote then qs:=not qs
        else fin:=not qs and (s[n]=ADelim);
        if not fin then inc(n);
        until (n>length(s)) or fin;
      if n<=3 then result:=''
      else Result:=AnsiDequotedStr(TrimRight(copy(s,1,pred(n))),AQuote);
      end
    else begin
      n:=pos(ADelim,s);
      if n=0 then n:=succ(length(s));
      Result:=copy(s,1,pred(n));
      end;
    delete(s,1,n);
    end
  else Result:='';
  end;

function MakeQuotedStr (const s : string; CheckChars : array of char) : string;
var
  i : integer;
  ok : boolean;
begin
  if (length(s)=0) then Result:=''
  else begin
    if length(CheckChars)=0 then ok:=true
    else begin
      ok:=false;
      for i:=Low(CheckChars) to High(CheckChars) do ok:=ok or (Pos(CheckChars[i],s)>0);
      end;
    if ok then Result:=AnsiQuotedStr(s,Quote) else Result:=s;
    end;
  end;

function CommaTextToLines (Commatext : string) : string;
begin
  Result:=ReadNxtQuotedStr(Commatext,Komma,Quote);
  while length(Commatext)>0 do
    Result:=Result+sLineBreak+ReadNxtQuotedStr(Commatext,Komma,Quote);
  end;

// Count number of lines in Text separated by sLineBreak
function CountLines (const Text : string) : integer;
var
  n : integer;
begin
  n:=0; Result:=1;
  repeat
    n:=PosEx(sLineBreak,Text,n+1);
    if n>0 then inc(Result);
    until (n=0) or (n>=length(Text));
  end;

// Calculate the maximum text width for multiline text
function MaxTextWidth(const Text : string; Canvas : TCanvas) : integer;
var
  n,k : integer;
begin
  n:=1; Result:=0;
  repeat
    k:=PosEx(sLineBreak,Text,n);
    if k=0 then k:=length(Text)+1;
    Result:=Max(Result,Canvas.TextWidth(copy(Text,n,k-n+1)));
    n:=k+length(sLineBreak);
    until (k=0) or (n>=length(Text));
  end;

function DuplicateQuotes(const S: string; Quote : Char): string;
var
  I : Integer;
begin
  Result:=S;
  for I := Length(Result) downto 1 do if Result[I]=Quote then Insert(Quote,Result,I);
  end;

{ ------------------------------------------------------------------- }
// Position of char in string starting at "Offset", ignore parts between quotes
function PosChar (const s: string; AChar,AQuote : char; Offset: Cardinal): Integer;
var
  i,ls : integer;
  qs   : boolean;
begin
  i:=Offset;
  ls:=length(s);
  qs:=true;
  while i<=ls do begin
    if s[i]=AQuote then qs:=not qs
    else if qs and (s[i]=AChar) then begin
      Result:=i;
      exit;
      end;
    inc(i);
    end;
  Result:=0;
  end;

{ ------------------------------------------------------------------- }
function GetPluralString (const sNo,sOne,sMany : string; n : integer) : string;
begin
  if n=1 then Result:=sOne
  else if (n=0) and (length(sNo)>0) then Result:=sNo
  else Result:=Format(sMany,[n]);
  end;

function GetPluralString (const sNo,sOne,sMany : string; n : integer; const s : string) : string;
begin
  if n=1 then Result:=Format(sOne,[s])
  else if (n=0) and (length(sNo)>0) then Result:=sNo
  else Result:=Format(sMany,[n,s]);
  end;

{ ------------------------------------------------------------------- }
(* Testen, ob gltige Integer oder Double Zahl in einem String s
   bis zum nchsten Begrenzer (Del) steht *)
function TestNxtNumber (s       : String;
                        Del     : char ) : boolean;
var
  n,i,ic : integer;
begin
  s:=RemSp(s); i:=pos (Del,s);
  if i=0 then i:=succ(length(s));
  val(copy(s,1,pred(i)),n,ic);
  TestNxtNumber:=ic=0;
  end;

{ ---------------------------------------------------------------- }
(* Dezimaltrenner auf Zeichen setzen, Standard = Punkt *)
procedure SetDecPtChar (Zeichen : char);
begin
  DezTr:=Zeichen;
  end;

{ ---------------------------------------------------------------- }
(* aktuellen Dezimaltrenner lesen *)
function GetDecPtChar : char;
begin
  GetDecPtChar:=DezTr;
  end;

{ ---------------------------------------------------------------- }
(* Tausendermarkierung, Standard = Leerzeichen *)
procedure ThMarkChar (Zeichen : char);
begin
  ThMark:=Zeichen;
  end;

{ ---------------------------------------------------------------- }
(* Integer-Zahl in String umsetzen *)
function StrInt (x : int64;
                 n : integer) : string;
(* n > 0 : Anzahl der Stellen
     = 0 : unformatiert
     < 0 : unformatiert mit Dreiergruppierung *)
var
  s   : string;
  i,j : integer;
begin
  s:=IntToStr(x);
  if n>length(s) then  s:=AddSp(s,n)
  else if n<0 then begin
    i:=succ(length(s));
    j:=i;
    while (j>4) do begin
      dec(j,3); insert (ThMark,s,j);
      end;
    end;
  StrInt:=s;
  end;

{ ---------------------------------------------------------------- }
(* Integer-Zahl in String mit fhrenden Nullen umsetzen *)
function ZStrInt (x : int64;
                  n : integer) : string;
var
  s : string;
begin
  s:=IntToStr(abs(x));
  if x<0 then begin
    s:=AddChar(s,'0',n-1);
    s:='-'+s;
    end
  else s:=AddChar(s,'0',n);
  ZStrInt:=s;
  end;

{ ---------------------------------------------------------------- }
function GroupDigits (s : string; Group : integer) : string;
var
  i : integer;
begin
  if (Group>0) then begin
    i:=length(s);
    while (i>Group) do begin
      dec(i,Group); Insert(' ',s,i+1);
      end;
    end;
  Result:=s;
  end;

{ ---------------------------------------------------------------- }
function NumberOfDigits (Value : int64) : integer;
begin
  if Value<0 then Result:=1 else Result:=0;
  Value:=abs(Value);
  repeat
    Value:=Value div 10;
    inc(Result);
    until Value=0;
  end;

{ ---------------------------------------------------------------- }
function IntToHexStr (Value,Digits,Group : cardinal) : string;
begin
  Result:=GroupDigits(IntToHex(Value,Digits),Group);
  end;

{ ---------------------------------------------------------------- }
function IntToDecimal(Value : int64; Digits,Group : cardinal) : string;
begin
  Result:=GroupDigits(ZStrInt(Value,Digits),Group);
  end;

{ ---------------------------------------------------------------- }
function IntToOctal (Value,Digits,Group : cardinal) : string;
begin
  Result:='';
  repeat
    Result:=chr((Value and 7)+48)+Result;
    Value:=Value div 8;
    until Value=0;
  Result:=GroupDigits(AddChar(Result,'0',Digits),Group);
  end;

{ ---------------------------------------------------------------- }
function IntToBin (Value,Digits,Group : cardinal) : string;
begin
  Result:='';
  repeat
    Result:=chr((Value and 1)+48)+Result;
    Value:=Value shr 1;
    until Value=0;
  Result:=GroupDigits(AddChar(Result,'0',Digits),Group);
  end;

{ ---------------------------------------------------------------- }
function TryHexStrToInt64(S : string; var val : int64) : boolean;
var
  minus : boolean;
begin
  val:=0; Result:=false;
  if length(s)>0 then begin
    minus:=s[1]='-';
    if minus then delete(s,1,1);
    if (length(s)>0) and (s[1]<>'$') then s:='$'+s;
    Result:=TryStrToInt64(DelSp(s),val);
    if Result and minus then val:=-val;
    end;
  end;

function TryHexStrToInt(const S : string; var val : integer) : boolean;
var
  n : int64;
begin
  n:=val; Result:=TryHexStrToInt64(S,n); val:=n;
  end;

function HexStrToInt (const S : string) : int64;
begin
  Result:=0;
  if not TryHexStrToInt64(S,Result) then
    raise EConvertError.Create(Format(rsHexError,[s]));
  end;

function TryOctalStrToInt64(S : string; var val : int64) : boolean;
var
  i     : integer;
  minus : boolean;
begin
  val:=0; Result:=false;
  if length(s)>0 then begin
    minus:=s[1]='-';
    if minus then delete(s,1,1);
    if (length(s)>0) and (s[1]='%') then delete(s,1,1);
    s:=DelSp(s);
    for i:=1 to length(s) do begin
      if (s[i]<#48) or (s[i]>#55) then exit
      else val:=8*val+ord(s[i])-48;
      end;
    if minus then val:=-val;
    Result:=true;
    end;
  end;

function TryOctalStrToInt(const S : string; var val : integer) : boolean;
var
  n : int64;
begin
  n:=val; Result:=TryOctalStrToInt64(S,n); val:=n;
  end;

function OctalStrToInt (const S: string) : int64;
begin
  Result:=0;
  if not TryOctalStrToInt64(s,Result) then
    raise EConvertError.Create(Format(rsOctError,[s]));
  end;

function TryBinStrToInt64(S: string; var val : int64) : boolean;
var
  i     : integer;
  minus : boolean;
begin
  val:=0; Result:=false;
  if length(s)>0 then begin
    minus:=s[1]='-';
    if minus then delete(s,1,1);
    if (length(s)>0) and (s[1]='!') then delete(s,1,1);
    s:=DelSp(s);
    for i:=1 to length(s) do begin
      if (s[i]<#48) or (s[i]>#49) then Exit
      else val:=2*val+ord(s[i])-48;
      end;
    if minus then val:=-val;
    Result:=true;
    end;
  end;

function TryBinStrToInt(const S : string; var val : integer) : boolean;
var
  n : int64;
begin
  n:=val; Result:=TryBinStrToInt64(S,n); val:=n;
  end;

function BinStrToInt (const S: string) : int64;
begin
  Result:=0;
  if not TryBinStrToInt64(s,Result) then
    raise EConvertError.Create(Format(rsBinError,[s]));
  end;

{ ---------------------------------------------------------------- }
(* Double-Zahl in String umsetzen *)
function StrDbl (x   : double;
                 n,d : integer) : string;
var
  s : string;
begin
  str (x:n:d,s);
  if DezTr<>DecPt then s:=ReplDecPt (s);
  StrDbl:=s;
  end;

function StrDbl (x   : double;
                 n,d : integer; DecPt : char) : string;
var
  dp : char;
begin
  dp:=DezTr; DezTr:=DecPt;
  Result:=StrDbl(x,n,d); DezTr:=dp;
  end;

{ ---------------------------------------------------------------- }
(* Double-Zahl in String mit fhrenden Nullen umsetzen *)
function ZStrDbl (x   : double;
                  n,d : integer) : string;
var
  s : string;
  i : integer;
begin
  str (abs(x):n:d,s);
  if x<0 then if s[1]=Space then s[1]:='-' else s:='-'+s;
  for i:=1 to n do if s[i]=' ' then s[i]:='0';
  if DezTr<>DecPt then s:=ReplDecPt(s);
  ZStrDbl:=s;
  end;

function ZStrDbl (x   : double;
                  n,d : integer; DecPt : char) : string;
var
  dp : char;
begin
  dp:=DezTr; DezTr:=DecPt;
  Result:=ZStrDbl(x,n,d); DezTr:=dp;
  end;

{ ------------------------------------------------------------------ }
(* Double-Zahl in String in Dreiergruppen umsetzen*)
function NStrDbl (x : double;
                  n,d : integer) : string;
(* Format wie StrDbl *)
var
  s   : string;
  i,j : integer;
begin
  s:=StrDbl(x,n,d);
  i:=pos(DezTr,s);
  if i=0 then i:=succ(length(s));
  j:=i;
  while (j>3) do begin
    dec(j,3); insert (ThMark,s,j);
    inc(i);
    end;
  inc(i,3);
  while (i<length(s)) do begin
    insert(ThMark,s,i); inc(i);
    end;
  NStrDbl:=s;
  end;

function NStrDbl (x   : double;
                  n,d : integer; DecPt : char) : string;
var
  dp : char;
begin
  dp:=DezTr; DezTr:=DecPt;
  Result:=NStrDbl(x,n,d); DezTr:=dp;
  end;

{ ---------------------------------------------------------------- }
// String in Fliekomma-Zahl mit Default-Wert und Punkt als Dezimaltrenner
function StringToFloat (const s : string; Default : extended) : extended;
var
  v : double;
begin
  if TryStrToFloat(SetDecSeps(s),v) then Result:=v else Result:=Default;
  end;

// Double-Zahl in String mit Punkt als Dezimaltrenner
function FloatToString (Value : extended; Format: TFloatFormat; Precision, Digits: Integer) : string;
begin
  Result:=ReplaceDecSeps(FloatToStrF(Value,Format,Precision,Digits));
  end;

{ ---------------------------------------------------------------- }
(* Bytes, je nach Gre in String umsetzen *)
function SizeToStr (Bytes : int64) : string;
var
  v : double;
  n : integer;
begin
  if Bytes<1024 then Result:=StrInt(Bytes,0)+' '+rsBytes  // bytes
  else begin
    v:=Bytes/1024;
    if v<1024 then begin
      if v<100 then n:=2 else n:=1;
      Result:=StrDbl(v,0,n)+' '+rsKBytes; // kBytes
      end
    else begin
      v:=v/1024;
      if v<1024 then begin
        if v<100 then n:=2 else n:=1;
        Result:=StrDbl(v,0,n)+' '+rsMBytes  // MBytes
        end
      else Result:=StrDbl(v/1024,0,2)+' '+rsGBytes; // GBytes
      end;
    end;
  end;

function SizeToStr (Bytes : int64; DecPt : char) : string;
var
  dp : char;
begin
  dp:=DezTr; DezTr:=DecPt;
  Result:=SizeToStr (Bytes); DezTr:=dp;
  end;

// String in Dateigre wandeln
function StrToSize (s : string; var size : int64) : boolean;
var
  f   : double;
  i,n : integer;
  c   : char;
  pf  : string;
begin
  Result:=false; size:=0;
  pf:=GetBytePrefixes;
  s:=Trim(s);
  if length(s)=0 then s:='0';
  if UpCase(s[length(s)])=rsBytes then delete(s,length(s),1);   // bytes
  s:=Trim(s);
  c:=s[length(s)];
  if (c>='0') then begin
    f:=1;
    if (c>'9') then begin
      n:=pos(' ',s);
      if n<=0 then Exit;
      i:=0;
      repeat
        f:=f*1024; inc(i);
        until (i>=length(pf)) or (c=pf[i]);
      if i>=length(pf) then Exit;
      s:=copy(s,1,n-1);
      end;
    try
      f:=StrToFloat(s)*f;
      size:=round(f);
    except
      Exit;
      end;
    Result:=true;
    end;
  end;

{ ---------------------------------------------------------------- }
// Umwandeln einer Menge in einen Hex-String und zurck
// Format des Strings: nnaabb.. (nn = Hexwert der nachfolgenden Bytes)
function SetToHex (ASet : PSetArray; Low,High : integer) : string;
var
  i : integer;
  b : byte;
begin
  b:=(High-Low) div 8+1;
  Result:=IntToHex(b,2);
  for i:=1 to b do Result:=Result+IntToHex(ASet^[i],2);
  end;

procedure HexToSet (const S : string; ASet: PSetArray);
var
  i,n : integer;
begin
  if (length(s)>=4) and TryStrToInt('$'+copy(s,1,2),n) then begin
    if length(s)>=2*(n+1) then begin
      if n>32 then n:=32;
      for i:=1 to n do if TryStrToInt('$'+copy(s,2*i+1,2),n) then ASet^[i]:=byte(n);
      end;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Wert auf Gltigkeit prfen
   true : x <> NoVal *)
function Valid (x,NoVal : double) : boolean;
begin
  Valid:=abs(x-NoVal)>0.1;
  end;

{ ---------------------------------------------------------------- }
(* Dezimalpunkt durch vorgegebenen Trenner (SetDecPtChar) ersetzen
   Kompatibilittsfunktion, fr Neuentwicklungen "SetDecSeps" (s.u.) *)
function ReplDecPt (s : string) : string;
var
  i : byte;
begin
  i:=pos(DecPt,s);
  if i>0 then s[i]:=DezTr;
  ReplDecPt:=s;
  end;

(* aktuellen Dezimaltrenner durch Punkt ersetzen
   Kompatibilittsfunktion, fr Neuentwicklungen "ReplaceDecSeps" (s.u.) *)
function SetDecPt (s : string) : string;
var
  i : byte;
begin
  i:=pos (DezTr,s);
  if i>0 then s[i]:=DecPt;
  SetDecPt:=s;
  end;

(* alle aktuellen Dezimaltrenner durch Punkt ersetzen *)
function ReplaceDecSeps (const s : string) : string;
begin
  if DecimalSeparator<>DecPt then Result:=ReplChars(s,DecimalSeparator,DecPt);
  end;

(* alle Punkte durch aktuellen Dezimlatrenner ersetzen *)
function SetDecSeps (const s : string) : string;
begin
  if DecimalSeparator<>DecPt then Result:=ReplChars(s,DecPt,DecimalSeparator);
  end;

{ ---------------------------------------------------------------- }
(* Cardinal in Hi- oder Lo-Wort zerlegen *)
function HiWord (n : cardinal) : word;
var
  Long : TLongWord absolute n;
begin
  HiWord:=Long.Hi;
  end;

function LoWord (n : cardinal) : word;
var
  Long : TLongWord absolute n;
begin
  LoWord:=Long.Lo;
  end;

(* Word aus Hi- und Lo-Byte zusammensetzen *)
function BytesToWord (HiByte,LoByte: byte) : word;
var
  Long : TWord;
begin
  with Long do begin
    Hi:=HiByte; Lo:=LoByte;
    Result:=Both;
    end;
  end;

(* Cardinal aus Hi- und Lo-Wort zusammensetzen *)
function WordsToCardinal (HiWord,LoWord : word) : cardinal;
var
  Long : TLongWord;
begin
  with Long do begin
    Hi:=HiWord; Lo:=LoWord;
    Result:=LongWord;
    end;
  end;

(* Cardinal aus 4 Bytes zusammensetzen *)
function BytesToCardinal (HHByte,HLByte,LHByte,LLByte : word) : cardinal;
var
  Long : TLongWord;
begin
  with Long do begin
    LoL:=LLbyte; LoH:=LHByte;
    HiL:=HLByte; HiH:=HHByte;
    Result:=LongWord;
    end;
  end;

(* Int64 aus Hi- und Lo-Cardinal zusammensetzen *)
function CardinalToInt64 (HiCard,LoCard : cardinal) : int64;
var
  Long : TInt64;
begin
  with Long do begin
    Hi:=HiCard; Lo:=LoCard;
    Result:=AsInt64;
    end;
  end;

{ ---------------------------------------------------------------- }
// Convert 16-bit oder 8-Bit number to hex string
function HexByte (b : byte) : AnsiString;
const
  HexTab : array [0..15] of AnsiChar = '0123456789ABCDEF';
begin
  HexByte:=HexTab[b shr 4] + HexTab[b and $f];
  end;

function HexWord (w : word) : AnsiString;
begin
  HexWord:=HexByte(hi(w))+HexByte(lo(w));
  end;

//-----------------------------------------------------------------------------
{ QUICKSORT sorts elements in the array "SList" of any record type of
  size "RSize" (in bytes) with indices between 0 and "Count"-1 (both inclusive).
  Note that the QuickSort procedure provides only an "interface" to the program.
  The actual processing takes place in the "Sort" procedure, which executes
  itself recursively.
  "Compare" is a callback function of type "TCompFunction" (see begin of unit).
  Algorithm based on Borland pascal sample "qsort.pas",
  modified J. Rathlev, Nov. 2001 (rathlev@physik.uni-kiel.de) }
//-----------------------------------------------------------------------------
type
  TByteDynArray = array of byte;

procedure QuickSort (var SList;
                     Count,RSize : Integer;                     Compare     : TCompFunction);
var
  {the  following variables are global to "Sort" to save memory in recursive calls }
  tmp,ns  : TByteDynArray;
  pl      : PByteArray;
  n,m     : integer;

  procedure Sort(nl,nr: Integer);
  var
    i,j   : integer;
  begin
    n:=RSize*((nl+nr) div 2);
    move (pl^[n],ns[0],RSize);
    i:=nl; j:=nr;
    repeat
      while Compare(pl^[RSize*i],ns[0]) do inc(i);
      while Compare(ns[0],pl^[RSize*j]) do dec(j);
      if i<=j then begin
        n:=RSize*i; m:=RSize*j;
        move (pl^[n],tmp[0],RSize);
        move (pl^[m],pl^[n],RSize);
        move (tmp[0],pl^[m],RSize);
        inc(i); dec(j);
        end;
      until i>j;
    if nl<j then Sort(nl,j);
    if i<nr then Sort(i,nr);
    end;

begin {QuickSort};
  SetLength(Tmp,RSize); SetLength(ns,RSize);
  pl:=@SList;
  Sort(0,Count-1);
  Tmp:=nil; ns:=nil;
  end;

{ ------------------------------------------------------------------- }
(* Erweiterte Druckerangaben *)
(* tatschliche Papierbreite in mm *)
function GetPaperWidth (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result:=round(25.4*GetDeviceCaps(Handle,PHYSICALWIDTH)/GetDeviceCaps(Handle,LOGPIXELSX));
  end;

(* tatschliche Papierhhe in mm *)
function GetPaperHeight (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result:=round(25.4*GetDeviceCaps(Handle,PHYSICALHEIGHT)/GetDeviceCaps(Handle,LOGPIXELSX));
  end;

(* nichtdruckbarer Bereich am linken Rand in mm *)
function GetLeftOffset (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result:=round(25.4*GetDeviceCaps(Handle,PHYSICALOFFSETX)/GetDeviceCaps(Handle,LOGPIXELSX));
  end;

(* nichtdruckbarer Bereich am oberen Rand in mm *)
function GetTopOffset (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result:=round(25.4*GetDeviceCaps(Handle,PHYSICALOFFSETY)/GetDeviceCaps(Handle,LOGPIXELSX));
  end;

(* nutzbare Papierbreite in mm *)
function GetMaxWidth (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result:=GetDeviceCaps(Handle, HorzSize);
  end;

(* nutzbare Papierhhe in mm *)
function GetMaxHeight (APrinter : TPrinter) : integer;
begin
  with APrinter do Result:=GetDeviceCaps(Handle, VertSize);
  Sleep(100);
  end;

// prfe, ob Duplex-Druck untersttzt wird
function SupportsDuplex (APrinter : TPrinter) : Boolean;
var
  Device,Driver,Port : array[0..255] of Char;
  hDevMode: THandle;
begin
  APrinter.GetPrinter(Device,Driver,Port,hDevmode);
  Result:=WinSpool.DeviceCapabilities(Device,Port,DC_DUPLEX,nil,nil)<>0;
  end;

// prfe,ob Duplex-Druck eingestellt ist
function UsesDuplex (APrinter : TPrinter) : Boolean;
var
  Device,Driver,Port : array[0..255] of Char;
  hDevMode: THandle;
  pDevmode: PDeviceMode;
begin
  Result:=false;
  APrinter.GetPrinter(Device,Driver,Port,hDevmode);
  if hDevmode<>0 then begin
     // lock it to get pointer to DEVMODE record
    pDevMode:=GlobalLock(hDevmode);
    if pDevmode<>nil then
      try
        Result:=pDevmode^.dmDuplex<>DMDUP_SIMPLEX;
      finally
        // unlock devmode handle.
        GlobalUnlock(hDevmode);
      end;
    end;
  end;

// Schalte Drucker auf Duplex
procedure SetToDuplex (APrinter : TPrinter);
var
  Device,Driver,Port : array[0..255] of Char;
  hDevMode: THandle;
  pDevmode: PDeviceMode;
begin
  APrinter.GetPrinter(Device,Driver,Port,hDevmode);
  if (hDevmode<>0) and (WinSpool.DeviceCapabilities(Device,Port,DC_DUPLEX,nil,nil)<>0) then begin
     // lock it to get pointer to DEVMODE record
    pDevMode:=GlobalLock(hDevmode);
    if pDevmode<>nil then
      try
        with pDevmode^ do begin
          dmDuplex:=DMDUP_VERTICAL;
          dmFields:=dmFields or DM_DUPLEX;
          end;
      finally
        // unlock devmode handle.
        GlobalUnlock(hDevmode);
      end;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Prfen, ob ein Fenster auf den Bildschirm passt, bei Bedarf
   Left, und Top anpassen
   an mehrere Monitore angepasst, Mrz. 2011 *)
procedure CheckScreenBounds (AScreen         : TScreen;
                             var ALeft,ATop : integer;
                             AWidth,AHeight : integer);
var
  mo : TMonitor;
begin
  with AScreen do begin
    mo:=MonitorFromPoint(Point(ALeft,ATop));
//    mo:=MonitorFromRect(Rect(ALeft,ATop,ALeft+AWidth,ATop+AHeight));
    with mo.WorkareaRect do begin
      if ALeft+AWidth>Right then ALeft:=Right-AWidth-20;
      if ALeft<Left then ALeft:=Left+20;
      if ATop+AHeight>Bottom then ATop:=Bottom-AHeight-30;
      if ATop<Top then ATop:=Top+20;
      end;
    end;
  end;

procedure FitToScreen (AScreen : TScreen; Control : TControl);
var
  il,it : integer;
begin
  with Control do begin
    il:=Left; it:=Top;
    CheckScreenBounds (AScreen,il,it,Width,Height);
    Left:=il; Top:=it;
    end;
  end;

// Adjust position of form to screen
procedure AdjustFormPosition (AScreen : TScreen; AForm : TForm;
          APos : TPoint; AtBottom : boolean = false);
begin
  with AForm,APos do begin
    if (Y < 0) or (X < 0) then Position:=poScreenCenter
    else begin
      Position:=poDesigned;
      if X<0 then X:=Left;
      if Y<0 then Y:=Top;
      if AtBottom then Y:=Y-Height;
      CheckScreenBounds(AScreen,x,y,Width,Height);  // DefaultMonitor = dmDesktop
      Left:=x; Top:=y;
      end;
    end;
  end;

{ ---------------------------------------------------------------- }
// neuer Message-Dialog mit Positionsprfung
// Delay = 0: ShowModal
//       > 0: Anzeigen und automatisch schlieen nach "Delay" in s
function MessageDialog(const Title,Msg: string; DlgType: TMsgDlgType;
                Buttons: TMsgDlgButtons; DefaultButton : TMsgDlgBtn;
                Pos : TPoint; Delay : integer;
                ADefaultMonitor : TDefaultMonitor = dmActiveForm) : integer;
begin
  with CreateMessageDialog(Msg,DlgType,Buttons,DefaultButton) do begin
    DefaultMonitor:=ADefaultMonitor;
//    ScaleBy(Screen.PixelsPerInch,PixelsPerInchOnDesign);
    Scaled:=true;
    try
      with Pos do begin
        if (Y < 0) and (X < 0) then Position:=poScreenCenter
        else begin
//          if X<0 then X:=Left;
//          if Y<0 then Y:=Top;
          CheckScreenBounds(Screen,x,y,Width,Height);
          Left:=x; Top:=y;
          end;
        end;
      if length(Title)>0 then Caption:=Title;
      FormStyle:=fsStayOnTop;
      if Delay<=0 then Result:=ShowModal
      else begin
        Show;
        Delay:=Delay*10;
        repeat
          Application.ProcessMessages;
          Sleep(100);
          dec(Delay);
          until (Delay=0) or (ModalResult<>mrNone);
        if ModalResult=mrNone then begin
          Close;
          Result:=mrOK;
          end
        else Result:=ModalResult;
        end;
    finally
      Free;
      end;
    end;
  end;

function MessageDialog(const Title,Msg: string; DlgType: TMsgDlgType;
                Buttons: TMsgDlgButtons;
                Pos : TPoint; Delay : integer;
                ADefaultMonitor : TDefaultMonitor = dmActiveForm) : integer;
var
  DefaultButton: TMsgDlgBtn;
begin
  if mbOk in Buttons then DefaultButton := mbOk else
    if mbYes in Buttons then DefaultButton := mbYes else
      DefaultButton := mbRetry;
  Result:=MessageDialog(Title,Msg,DlgType,Buttons,DefaultButton,Pos,Delay,ADefaultMonitor);
end;

function MessageDialog(const Title,Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons) : integer;
begin
  Result:=MessageDialog(Title,Msg,DlgType,Buttons,CenterPos,0);
  end;

function MessageDialog(const Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons) : integer;
begin
  Result:=MessageDialog('',Msg,DlgType,Buttons,CenterPos,0);
  end;

function MessageDialog(Pos : TPoint; const Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons) : integer;
begin
  Result:=MessageDialog('',Msg,DlgType,Buttons,Pos,0);
  end;

{ ---------------------------------------------------------------- }
// Besttigung in Bildschirmmitte (X<0) oder an Position X,Y
function ConfirmDialog (const Title,Msg : string; Pos : TPoint) : boolean;
begin
  Result:=MessageDialog (Title,Msg,mtConfirmation,[mbYes,mbNo],Pos,0)=mrYes;
  end;

// Besttigung auf einstellbarem Monitor
function ConfirmDialog (Pos : TPoint; const Msg : string; DefaultButton : TMsgDlgBtn = mbYes;
                        ADefaultMonitor : TDefaultMonitor = dmActiveForm) : boolean;
begin
  Result:=MessageDialog ('',Msg,mtConfirmation,[mbYes,mbNo],DefaultButton,Pos,0,ADefaultMonitor)=mrYes;
  end;

// Besttigung in Bildschirmmitte
function ConfirmDialog (const Title,Msg  : string) : boolean;
begin
  Result:=ConfirmDialog(Title,Msg,CenterPos);
  end;

function ConfirmDialog (const Msg : string; DefaultButton : TMsgDlgBtn) : boolean;
begin
  Result:=MessageDialog ('',Msg,mtConfirmation,[mbYes,mbNo],DefaultButton,CenterPos,0)=mrYes;
  end;

// Information an Position ausgeben
procedure InfoDialog (const Title,Msg : string; Pos : TPoint);
begin
  MessageDialog (Title,Msg,mtInformation,[mbOK],Pos,0);
  end;

procedure InfoDialog (Pos : TPoint; const Msg :string);
begin
  InfoDialog('',Msg,Pos);
  end;

// Information in Bildschirmmitte ausgeben
procedure InfoDialog (const Title,Msg :string);
begin
  InfoDialog(Title,Msg,CenterPos);
  end;

// Information in Bildschirmmitte ausgeben und fr Delay s anzeigen
procedure InfoDialog (const Title,Msg : string; Delay : integer); overload;
begin
  MessageDialog (Title,Msg,mtInformation,[mbOK],CenterPos,Delay);
  end;

procedure InfoDialog (const Msg :string);
begin
  InfoDialog('',Msg,CenterPos);
  end;

// Fehlermeldung an Position ausgeben
procedure ErrorDialog (const Title,Msg : string; Pos : TPoint);
begin
  MessageDialog (Title,Msg,mtError,[mbOK],Pos,0);
  end;

procedure ErrorDialog (const Title,Msg : string; x,y : integer);
begin
  MessageDialog (Title,Msg,mtError,[mbOK],Point(x,y),0);
  end;

procedure ErrorDialog (Pos : TPoint; const Msg : string);
begin
  ErrorDialog('',Msg,Pos);
  end;

// Fehlermeldung in Bildschirmmitte ausgeben und fr Delay s anzeigen
procedure ErrorDialog (const Title,Msg : string; Delay : integer); overload;
begin
  MessageDialog (Title,Msg,mtError,[mbOK],CenterPos,Delay);
  end;

// Fehlermeldung in Bildschirmmitte ausgeben
procedure ErrorDialog (const Title,Msg : string);
begin
  ErrorDialog(Title,Msg,CenterPos);
  end;

procedure ErrorDialog (const Msg : string);
begin
  ErrorDialog('',Msg,CenterPos);
  end;

{ ------------------------------------------------------------------- }
// get current cursor position
function CursorPos : TPoint;
begin
  GetCursorPos(Result);
  end;

{ ------------------------------------------------------------------- }
// Add offset to point
function AddOffsetPos(Pos1,Pos2 : TPoint) : TPoint;
begin
  with Result do begin
    x:=Pos1.x+Pos2.x; y:=Pos1.y+Pos2.y;
    end;
  end;

// get current cursor position, add Offset
function CursorPos (Offset : TPoint): TPoint;
var
  pos : TPoint;
begin
  GetCursorPos(pos);
  Result:=AddOffsetPos(pos,Offset);
  end;

{ ------------------------------------------------------------------- }
// position of component
function TopLeftPos (AControl : TControl; Offset : TPoint) : TPoint;
begin
  with AControl do if assigned(Parent) then Result:=Parent.ClientToScreen(Point(Left,Top))
  else Result:=Point(Left,Top);
  Result:=AddOffsetPos(Result,Offset);
  end;

function TopLeftPos (AControl : TControl) : TPoint;
begin
  Result:=TopLeftPos (AControl,Point(0,0));
  end;

function BottomLeftPos (AControl : TControl; Offset : TPoint) : TPoint;
begin
  with AControl do if assigned(Parent) then Result:=Parent.ClientToScreen(Point(Left,Top+Height))
  else Result:=Point(Left,Top+Height);
  Result:=AddOffsetPos(Result,Offset);
  end;

function BottomLeftPos (AControl : TControl) : TPoint;
begin
  Result:=BottomLeftPos (AControl,Point(0,0));
  end;

function TopRightPos (AControl : TControl; Offset : TPoint) : TPoint;
begin
  with AControl do if assigned(Parent) then Result:=Parent.ClientToScreen(Point(Left+Width,Top))
  else Result:=Point(Left+Width,Top);
  Result:=AddOffsetPos(Result,Offset);
  end;

function TopRightPos (AControl : TControl) : TPoint;
begin
  Result:=TopRightPos (AControl,Point(0,0));
  end;

function BottomRightPos (AControl : TControl; Offset : TPoint) : TPoint;
begin
  with AControl do if assigned(Parent) then Result:=Parent.ClientToScreen(Point(Left+Width,Top+Height))
  else Result:=Point(Left+Width,Top+Height);
  Result:=AddOffsetPos(Result,Offset);
  end;

function BottomRightPos (AControl : TControl) : TPoint;
begin
  Result:=BottomRightPos (AControl,Point(0,0));
  end;

// area of component
function GetRect (AControl : TControl) : TRect;
begin
  with AControl do Result:=Rect(Left,Top,Left+Width,Top+Height);
  end;

{ ------------------------------------------------------------------- }
// History list management
const
  iniHist = 'History';

procedure LoadHistory (IniFile : TIniFile; Section,Ident : string;
                       History : TStrings; MaxCount : integer);
var
  i : integer;
  s : string;
begin
  with IniFile do begin
    if SectionExists(Section) then begin
      if length(Ident)=0 then Ident:=iniHist;
      History.Clear;
      for i:=0 to MaxCount-1 do begin
        s:=ReadString(Section,Ident+IntToStr(i),'');
        if s<>'' then History.Add(s);
        end;
      end;
    end;
  end;

procedure LoadHistory (IniFile : TIniFile; Section,Ident : string;
                       History : TStrings);
begin
  LoadHistory(IniFile,Section,Ident,History,MaxHist);
  end;

procedure LoadHistory (IniName : string; Section,Ident : string;
                       History : TStrings; MaxCount : integer);
var
  IniFile : TIniFile;
begin
  IniFile:=TIniFile.Create(IniName);
  LoadHistory(IniFile,Section,Ident,History,MaxCount);
  IniFile.Free;
  end;

procedure LoadHistory (IniName : string; Section,Ident : string;
                       History : TStrings); overload;
begin
  LoadHistory(IniName,Section,Ident,History,MaxHist);
  end;

procedure SaveHistory (IniFile : TIniFile; Section,Ident : string;
                       Erase : boolean; History : TStrings; MaxCount : integer);
var
  i,n : integer;
begin
  with IniFile do begin
    if length(Ident)=0 then Ident:=iniHist;
    if Erase then EraseSection (Section);
    with History do begin
      if Count>MaxCount then n:=MaxCount else n:=Count;
      for i:=0 to n-1 do WriteString(Section,Ident+IntToStr(i),Strings[i]);
      end;
    end;
  end;

procedure SaveHistory (IniFile : TIniFile; Section,Ident : string;
                       Erase : boolean; History : TStrings);
begin
  SaveHistory(IniFile,Section,Ident,Erase,History,MaxHist);
  end;

procedure SaveHistory (IniName: string; Section,Ident : string;
                       Erase : boolean; History : TStrings; MaxCount : integer);
var
  IniFile : TIniFile;
begin
  IniFile:=TIniFile.Create(IniName);
  SaveHistory(IniFile,Section,Ident,Erase,History,MaxHist);
  IniFile.Free;
  end;

procedure SaveHistory (IniName: string; Section,Ident : string;
                       Erase : boolean; History : TStrings);
begin
  SaveHistory(IniName,Section,Ident,Erase,History,MaxHist);
  end;

// move or add item "hs" to begin of history list
procedure AddToHistory (History : TStrings; hs : string; MaxCount : integer);
var
  n : integer;
begin
  if length(hs)>0 then with History do begin
    n:=IndexOf(hs);
    if n<0 then begin
      if Count>=MaxCount then Delete (Count-1);
      Insert (0,hs);
      end
    else if n>0 then Move (n,0);
    end;
  end;

procedure AddToHistory (History : TStrings; hs : string);
begin
  AddToHistory (History,hs,MaxHist);
  end;

procedure RemoveFromHistory (History : TStrings; hs : string);
var
  n : integer;
begin
  if length(hs)>0 then with History do begin
    n:=IndexOf(hs);
    if n>=0 then Delete (n);
    end;
  end;

//-----------------------------------------------------------------------------
procedure FreeListObjects (Liste : TStrings);
var
  i : integer;
begin
  with Liste do begin
    for i:=0 to Count-1 do if assigned(Objects[i]) then begin
      Objects[i].Free; Objects[i]:=nil;
      end;
    end;
  end;

procedure FreeListViewData (Liste : TListItems);
var
  i : integer;
begin
  with Liste do for i:=0 to Count-1 do with Item[i] do if Data<>nil then begin
    TObject(Data).Free; Data:=nil;
    end;
  end;

//-----------------------------------------------------------------------------
// Listview-Index aus Caption ermitteln (wie IndexOf bei TListBox)
function GetListViewIndex (lv : TListView; ACaption : string): integer;
begin
  with lv.Items do for Result:=0 to Count-1 do
    if AnsiSameText(Item[Result].Caption,ACaption) then Exit;
  Result:=-1;
  end;

// Subitem-Index aus der Mausposition ermitteln (nur vsReport)
function GetColumnIndexAt (ListView : TListView; Pos : integer) : integer;
var
  x : integer;
begin
  with ListView.Columns do begin
    x:=0;
    for Result:=0 to Count-1 do with Items[Result] do begin
      if (Pos>=x) and (Pos<x+Width) then Exit;
      x:=x+Width;
      end;
    end;
  Result:=-1;
  end;

// TopItem auf Index setzen (nur vsReport)
procedure SetListViewTopItem (lv : TListView; AIndex : integer; Select : boolean);
var
  n : integer;
begin
  with lv do if (AIndex>=0) and (Items.Count>0) and (AIndex<Items.Count) then begin
    with TopItem.DisplayRect(drBounds)do n:=Top-Bottom;
    Scroll(0,n*(TopItem.Index-AIndex));
    if Select then ItemIndex:=AIndex;
    end;
  end;

//-----------------------------------------------------------------------------
// Tastaturpuffer lschen
function ClearKeyboardBuffer : Integer;
var
   Msg: TMsg;
begin
  Result := 0;
  while PeekMessage(Msg,0,WM_KEYFIRST,WM_KEYLAST,PM_REMOVE) do inc(Result);
  end;

{ ------------------------------------------------------------------- }
(* Verzgerungszeit unter Verwendung der Windows Waitable Timer Objects
   Wait: warten bis Zeit in ms abgelaufen ist (min. Zeit 100 ns) *)

constructor TWaitTimer.Create;
begin
  inherited Create;
  FTimer:=CreateWaitableTimer(nil,true,'WaitableTimer');
  end;

destructor TWaitTimer.Destroy;
begin
  CloseHandle(FTimer);
  end;

procedure TWaitTimer.Wait (DelayTime : double); // Time in ms
var
  d : TLargeInteger;
begin
  d:=-round(10000*DelayTime);
  SetWaitableTimer(FTimer,d,0,nil,nil,false);
  repeat until WaitForSingleObject(FTimer,INFINITE)=WAIT_OBJECT_0;
  end;

{ ------------------------------------------------------------------- }
// Anzeige eines Hinweisfensters (THintWindow), das nach einstellbarer Zeit (Delay)
// automatisch verschwindet
constructor TTimerHint.Create (AOwner: TComponent; Delay : integer);
begin
  inherited Create(AOwner);
  FTimer:=TTimer.Create(AOwner);
  with FTimer do begin
    Interval:=Delay;
    Enabled:=false;
    OnTimer:=Terminate;
    end;
  FOnTerminate:=nil;
  end;

procedure TTimerHint.CreateParams(var Params: TCreateParams);
begin
  inherited;
  Params.WndParent := 0;
  end;

destructor TTimerHint.Destroy;
begin
  FTimer.Free;
  inherited Destroy;
  end;

procedure TTimerHint.ShowHint (r : TRect; AHint : string);
begin
  ActivateHint(r,AHint);
  FTimer.Enabled:=true;
  end;

procedure TTimerHint.Terminate (Sender : TObject);
begin
  FTimer.Enabled:=false;
  ReleaseHandle;
  if assigned(FOnTerminate) then FOnTerminate(self);
  end;

procedure TTimerHint.HideHint;
begin
  FTimer.Enabled:=false;
  ReleaseHandle;
  end;

{ ------------------------------------------------------------------- }
begin
  DezTr:=DecimalSeparator;
  ThMark:=Space;   // Zeichen fr Tausendergruppen
  TmMark:=DPunkt;  // Zeichen fr Uhrzeitgruppen hh:mm:ss
end.

